@@ -50,70 +50,74 @@
 				       (elmo-make-folder
 					(elmo-match-string 3 name)))
     (elmo-pipe-folder-set-copy-internal folder
-					(string= ":" (elmo-match-string 2 name))))
+					(string= ":"
+						 (elmo-match-string 2 name))))
   folder)
 
 (luna-define-method elmo-folder-get-primitive-list ((folder elmo-pipe-folder))
-  (elmo-flatten
-   (mapcar
-    'elmo-folder-get-primitive-list
-    (list (elmo-pipe-folder-src-internal folder)
-	  (elmo-pipe-folder-dst-internal folder)))))
+  (nconc
+   (elmo-folder-get-primitive-list (elmo-pipe-folder-src-internal folder))
+   (elmo-folder-get-primitive-list (elmo-pipe-folder-dst-internal folder))))
 
 (luna-define-method elmo-folder-contains-type ((folder elmo-pipe-folder)
 					       type)
   (or (elmo-folder-contains-type (elmo-pipe-folder-src-internal folder) type)
       (elmo-folder-contains-type (elmo-pipe-folder-dst-internal folder) type)))
 
-(luna-define-method elmo-folder-msgdb-create ((folder elmo-pipe-folder)
-					      numlist new-mark already-mark
-					      seen-mark important-mark
-					      seen-list)
-  (elmo-folder-msgdb-create (elmo-pipe-folder-dst-internal folder)
-			    numlist new-mark already-mark
-			    seen-mark important-mark seen-list))
-
 (luna-define-method elmo-folder-append-messages ((folder elmo-pipe-folder)
 						 src-folder numbers
-						 unread-marks
 						 &optional same-number)
   (elmo-folder-append-messages (elmo-pipe-folder-dst-internal folder)
 			       src-folder numbers
-			       unread-marks
 			       same-number))
 
 (luna-define-method elmo-folder-append-buffer ((folder elmo-pipe-folder)
-					       unread &optional number)
+					       &optional flag number)
   (elmo-folder-append-buffer (elmo-pipe-folder-dst-internal folder)
-			     unread number))
+			     flag number))
 
 (luna-define-method elmo-message-fetch ((folder elmo-pipe-folder)
 					number strategy
-					&optional section outbuf unseen)
-  (elmo-message-fetch (elmo-pipe-folder-dst-internal folder)
-		      number strategy section outbuf unseen))
+					&optional unseen section)
+  (when (elmo-message-fetch (elmo-pipe-folder-dst-internal folder)
+			    number strategy unseen section)
+    (unless unseen
+      (elmo-folder-notify-event folder 'flag-changed (list number)))
+    t))
+
+(luna-define-method elmo-folder-clear :after ((folder elmo-pipe-folder)
+					      &optional keep-killed)
+  (unless keep-killed
+    (elmo-pipe-folder-copied-list-save folder nil)))
 
 (luna-define-method elmo-folder-delete-messages ((folder elmo-pipe-folder)
 						 numbers)
   (elmo-folder-delete-messages (elmo-pipe-folder-dst-internal folder)
 			       numbers))
 
+(luna-define-method elmo-folder-detach-messages ((folder elmo-pipe-folder)
+						 numbers)
+  (elmo-folder-detach-messages (elmo-pipe-folder-dst-internal folder)
+			       numbers))
+
 (defvar elmo-pipe-drained-hook nil "A hook called when the pipe is flushed.")
 
 (defsubst elmo-pipe-folder-list-target-messages (src &optional ignore-list)
   (unwind-protect
       (progn
 	(elmo-folder-set-killed-list-internal src ignore-list)
-	(elmo-folder-list-messages src))
+	(elmo-folder-list-messages src t))
     (elmo-folder-set-killed-list-internal src nil)))
 
 (defun elmo-pipe-drain (src dst &optional copy ignore-list)
   "Move or copy all messages of SRC to DST."
-  (let ((elmo-inhibit-number-mapping (not copy)) ; No need to use UIDL
+  (let ((elmo-inhibit-number-mapping (and (eq (elmo-folder-type-internal
+					       src) 'pop3)
+					  (not copy))) ; No need to use UIDL
 	msgs len)
     (message "Checking %s..." (elmo-folder-name-internal src))
     ;; Warnnig: some function requires msgdb
-    ;;  but elmo-folder-open-internal do not load msgdb.
+    ;; but elmo-folder-open-internal do not load msgdb.
     (elmo-folder-open-internal src)
     (setq msgs (elmo-pipe-folder-list-target-messages src ignore-list)
 	  len (length msgs))
@@ -124,14 +128,11 @@
 			     "Copying messages..."
 			   "Moving messages...")))
     (unwind-protect
-	(elmo-folder-move-messages src msgs dst
-				   nil nil copy)
+	(elmo-folder-move-messages src msgs dst copy)
       (elmo-progress-clear 'elmo-folder-move-messages))
     (when (and copy msgs)
       (setq ignore-list (elmo-number-set-append-list ignore-list
 						     msgs)))
-    ;; Don't save msgdb here.
-    ;; Because summary view of original folder is not updated yet.
     (elmo-folder-close-internal src)
     (run-hooks 'elmo-pipe-drained-hook)
     ignore-list))
@@ -154,40 +155,22 @@
 		      (expand-file-name "pipe" elmo-msgdb-directory)))
    copied-list))
 
+(luna-define-method elmo-folder-msgdb ((folder elmo-pipe-folder))
+  (elmo-folder-msgdb (elmo-pipe-folder-dst-internal folder)))
+
 (luna-define-method elmo-folder-open-internal ((folder elmo-pipe-folder))
-  (elmo-folder-open-internal (elmo-pipe-folder-dst-internal folder))
-  (let ((src-folder (elmo-pipe-folder-src-internal folder))
-	(dst-folder (elmo-pipe-folder-dst-internal folder)))
-    (when (and (elmo-folder-plugged-p src-folder)
-	       (elmo-folder-plugged-p dst-folder))
-      (if (elmo-pipe-folder-copy-internal folder)
-	  (elmo-pipe-folder-copied-list-save
-	   folder
-	   (elmo-pipe-drain src-folder
-			    dst-folder
-			    'copy
-			    (elmo-pipe-folder-copied-list-load folder)))
-	(elmo-pipe-drain src-folder dst-folder)))))
+  (elmo-folder-open-internal (elmo-pipe-folder-dst-internal folder)))
 
 (luna-define-method elmo-folder-close-internal ((folder elmo-pipe-folder))
   (elmo-folder-close-internal(elmo-pipe-folder-dst-internal folder)))
 
-(luna-define-method elmo-folder-list-messages-internal
-  ((folder elmo-pipe-folder) &optional nohide)
-  (elmo-folder-list-messages-internal (elmo-pipe-folder-dst-internal
-				       folder) nohide))
+(luna-define-method elmo-folder-list-messages ((folder elmo-pipe-folder)
+					       &optional visible-only in-msgdb)
+  ;; Use target folder's killed-list in the pipe folder.
+  (elmo-folder-list-messages (elmo-pipe-folder-dst-internal
+			      folder) visible-only in-msgdb))
 
-(luna-define-method elmo-folder-list-unreads-internal
-  ((folder elmo-pipe-folder) unread-marks &optional mark-alist)
-  (elmo-folder-list-unreads-internal (elmo-pipe-folder-dst-internal folder)
-				     unread-marks mark-alist))
-
-(luna-define-method elmo-folder-list-importants-internal
-  ((folder elmo-pipe-folder) important-mark)
-  (elmo-folder-list-importants-internal (elmo-pipe-folder-dst-internal folder)
-					important-mark))
-
-(luna-define-method elmo-folder-status ((folder elmo-pipe-folder))
+(luna-define-method elmo-folder-diff ((folder elmo-pipe-folder))
   (elmo-folder-open-internal (elmo-pipe-folder-src-internal folder))
   (elmo-folder-open-internal (elmo-pipe-folder-dst-internal folder))
   (let* ((elmo-inhibit-number-mapping
@@ -195,10 +178,14 @@
 	 (src-length (length (elmo-pipe-folder-list-target-messages
 			      (elmo-pipe-folder-src-internal folder)
 			      (elmo-pipe-folder-copied-list-load folder))))
-	 (dst-list (elmo-folder-list-messages
-		    (elmo-pipe-folder-dst-internal folder))))
-    (prog1 (cons (+ src-length (elmo-max-of-list dst-list))
-		 (+ src-length (length dst-list)))
+	 (dst-diff (elmo-folder-diff (elmo-pipe-folder-dst-internal folder))))
+    (prog1
+	(cond
+	 ((consp (cdr dst-diff)) ; new unread all
+	  (mapcar (lambda (number) (+ number src-length)) dst-diff))
+	 (t
+	  (cons (+ (or (car dst-diff) 0) src-length)
+		(+ (or (cdr dst-diff) 0) src-length))))
       ;; No save.
       (elmo-folder-close-internal (elmo-pipe-folder-src-internal folder))
       (elmo-folder-close-internal (elmo-pipe-folder-dst-internal folder)))))
@@ -243,8 +230,7 @@
   (elmo-message-use-cache-p (elmo-pipe-folder-dst-internal folder) number))
 
 (luna-define-method elmo-folder-check ((folder elmo-pipe-folder))
-  (elmo-folder-close-internal folder)
-  (elmo-folder-open-internal folder))
+  (elmo-folder-check (elmo-pipe-folder-dst-internal folder)))
 
 (luna-define-method elmo-folder-plugged-p ((folder elmo-pipe-folder))
   (and (elmo-folder-plugged-p (elmo-pipe-folder-src-internal folder))
@@ -281,29 +267,24 @@
   (elmo-folder-message-make-temp-files
    (elmo-pipe-folder-dst-internal folder) numbers start-number))
 
-(luna-define-method elmo-folder-mark-as-read ((folder elmo-pipe-folder)
-					      numbers)
-  (elmo-folder-mark-as-read (elmo-pipe-folder-dst-internal folder)
-			    numbers))
+(luna-define-method elmo-folder-set-flag ((folder elmo-pipe-folder)
+					  numbers
+					  flag
+					  &optional is-local)
+  (elmo-folder-set-flag (elmo-pipe-folder-dst-internal folder)
+			numbers flag is-local)
+  (elmo-folder-notify-event folder 'flag-changed numbers))
 
-(luna-define-method elmo-folder-unmark-read ((folder elmo-pipe-folder)
-					      numbers)
-  (elmo-folder-unmark-read (elmo-pipe-folder-dst-internal folder)
-			   numbers))
+(luna-define-method elmo-folder-unset-flag ((folder elmo-pipe-folder)
+					    numbers
+					    flag
+					    &optional is-local)
+  (elmo-folder-unset-flag (elmo-pipe-folder-dst-internal folder)
+			  numbers flag is-local)
+  (elmo-folder-notify-event folder 'flag-changed numbers))
 
-(luna-define-method elmo-folder-unmark-important ((folder elmo-pipe-folder)
-						  numbers)
-  (elmo-folder-unmark-important (elmo-pipe-folder-dst-internal folder)
-				numbers))
-
-(luna-define-method elmo-folder-mark-as-important ((folder elmo-pipe-folder)
-						   numbers)
-  (elmo-folder-mark-as-important (elmo-pipe-folder-dst-internal folder)
-				 numbers))
-
-;; disabled since msgdb of dst-folder is not synchronized [wl:11272]
-;(luna-define-method elmo-folder-pack-numbers ((folder elmo-pipe-folder))
-;  (elmo-folder-pack-numbers (elmo-pipe-folder-dst-internal folder)))
+(luna-define-method elmo-folder-pack-numbers ((folder elmo-pipe-folder))
+  (elmo-folder-pack-numbers (elmo-pipe-folder-dst-internal folder)))
 
 (luna-define-method elmo-folder-rename ((folder elmo-pipe-folder) new-name)
   (let* ((new-folder (elmo-make-folder new-name)))
@@ -327,6 +308,85 @@
 		      (elmo-pipe-folder-dst-internal new-folder))
     (elmo-msgdb-rename-path folder new-folder)))
 
+(luna-define-method elmo-folder-synchronize ((folder elmo-pipe-folder)
+					     &optional
+					     disable-killed
+					     ignore-msgdb
+					     no-check
+					     mask)
+  (let ((src-folder (elmo-pipe-folder-src-internal folder))
+	(dst-folder (elmo-pipe-folder-dst-internal folder)))
+    (when (and (elmo-folder-plugged-p src-folder)
+	       (elmo-folder-plugged-p dst-folder))
+      (if (elmo-pipe-folder-copy-internal folder)
+	  (elmo-pipe-folder-copied-list-save
+	   folder
+	   (elmo-pipe-drain src-folder
+			    dst-folder
+			    'copy
+			    (elmo-pipe-folder-copied-list-load folder)))
+	(elmo-pipe-drain src-folder dst-folder))))
+  (elmo-folder-synchronize
+   (elmo-pipe-folder-dst-internal folder)
+   disable-killed ignore-msgdb no-check mask))
+
+(luna-define-method elmo-folder-list-flagged ((folder elmo-pipe-folder)
+					      flag
+					      &optional in-msgdb)
+  (elmo-folder-list-flagged
+   (elmo-pipe-folder-dst-internal folder) flag in-msgdb))
+
+(luna-define-method elmo-folder-commit ((folder elmo-pipe-folder))
+  (elmo-folder-commit (elmo-pipe-folder-dst-internal folder)))
+
+(luna-define-method elmo-folder-length ((folder elmo-pipe-folder))
+  (elmo-folder-length (elmo-pipe-folder-dst-internal folder)))
+
+(luna-define-method elmo-message-flag-available-p ((folder elmo-pipe-folder)
+						   number flag)
+  (elmo-message-flag-available-p
+   (elmo-pipe-folder-dst-internal folder)
+   number flag))
+
+(luna-define-method elmo-folder-count-flags ((folder elmo-pipe-folder))
+  (elmo-folder-count-flags (elmo-pipe-folder-dst-internal folder)))
+
+(luna-define-method elmo-message-flags ((folder elmo-pipe-folder) number)
+  (elmo-message-flags (elmo-pipe-folder-dst-internal folder) number))
+
+(luna-define-method elmo-message-field ((folder elmo-pipe-folder)
+					number field)
+  (elmo-message-field (elmo-pipe-folder-dst-internal folder)
+		      number
+		      field))
+
+(luna-define-method elmo-message-set-cached ((folder elmo-pipe-folder)
+					     number cached)
+  (elmo-message-set-cached (elmo-pipe-folder-dst-internal folder)
+			   number cached))
+
+(luna-define-method elmo-find-fetch-strategy ((folder elmo-pipe-folder)
+					      number
+					      &optional
+					      ignore-cache
+					      require-entireness)
+  (elmo-find-fetch-strategy (elmo-pipe-folder-dst-internal folder)
+			    number
+			    ignore-cache
+			    require-entireness))
+
+(luna-define-method elmo-message-number ((folder elmo-pipe-folder)
+					 message-id)
+  (elmo-message-number (elmo-pipe-folder-dst-internal folder)
+		       message-id))
+
+(luna-define-method elmo-message-entity ((folder elmo-pipe-folder) key)
+  (elmo-message-entity (elmo-pipe-folder-dst-internal folder) key))
+
+(luna-define-method elmo-message-folder ((folder elmo-pipe-folder)
+					 number)
+  (elmo-message-folder (elmo-pipe-folder-dst-internal folder) number))
+
 (require 'product)
 (product-provide (provide 'elmo-pipe) (require 'elmo-version))
 
