@@ -28,11 +28,16 @@
 
 ;;; Code:
 ;;
-
 (require 'mime-view)
 (require 'mime-edit)
 (require 'mime-play)
 (require 'elmo)
+(require 'wl-vars)
+
+(eval-when-compile
+  (defalias-maybe 'pgg-decrypt-region 'ignore)
+  (defalias-maybe 'pgg-display-output-buffer 'ignore)
+  (defalias-maybe 'pgg-verify-region 'ignore))
 
 ;;; Draft
 
@@ -54,15 +59,56 @@ has Non-nil value\)"
 		     (function wl-draft-yank-to-draft-buffer))))
 	(mime-preview-following-method-alist
 	 (list (cons 'wl-original-message-mode
-		     (function wl-draft-yank-to-draft-buffer)))))
-    (if (get-buffer (wl-current-message-buffer))
+		     (function wl-draft-yank-to-draft-buffer))))
+	(message-buffer (wl-current-message-buffer)))
+    (if message-buffer
 	(save-excursion
-	  (set-buffer (wl-current-message-buffer))
+	  (set-buffer message-buffer)
 	  (save-restriction
 	    (widen)
-	    (if (wl-region-exists-p)
-		(wl-mime-preview-follow-current-region)
-	      (mime-preview-follow-current-entity)))))))
+	    (cond
+	     ((wl-region-exists-p)
+	      (wl-mime-preview-follow-current-region))
+	     ((not (wl-message-mime-analysis-p
+		    (wl-message-buffer-display-type)))
+	      (wl-mime-preview-follow-no-mime
+	       (wl-message-buffer-display-type)))
+	     (t
+	      (mime-preview-follow-current-entity)))))
+      (error "No message."))))
+
+;; modified mime-preview-follow-current-entity from mime-view.el
+(defun wl-mime-preview-follow-no-mime (display-type)
+  "Write follow message to current message, without mime.
+It calls following-method selected from variable
+`mime-preview-following-method-alist'."
+  (interactive)
+  (let* ((mode (mime-preview-original-major-mode 'recursive))
+	 (new-name (format "%s-no-mime" (buffer-name)))
+	 new-buf min beg end
+	 (entity (get-text-property (point-min) 'elmo-as-is-entity))
+	 (the-buf (current-buffer))
+	 fields)
+    (save-excursion
+      (goto-char (point-min))
+      (setq min (point-min)
+	    beg (re-search-forward "^$" nil t)
+	    end (point-max)))
+    (save-excursion
+      (set-buffer (setq new-buf (get-buffer-create new-name)))
+      (erase-buffer)
+      (insert-buffer-substring the-buf beg end)
+      (goto-char (point-min))
+      ;; Insert all headers.
+      (let ((elmo-mime-display-header-analysis
+	     (wl-message-mime-analysis-p display-type 'header)))
+	(elmo-mime-insert-sorted-header entity))
+      (let ((f (cdr (assq mode mime-preview-following-method-alist))))
+	(if (functionp f)
+	    (funcall f new-buf)
+	  (message
+	   "Sorry, following method for %s is not implemented yet."
+	   mode))))))
 
 ;; modified mime-preview-follow-current-entity from mime-view.el
 (defun wl-mime-preview-follow-current-region ()
@@ -87,7 +133,8 @@ It calls following-method selected from variable
 	(insert-buffer-substring the-buf r-beg r-end)
 	(goto-char (point-min))
 	(let ((current-entity
-	       (if (and (eq (mime-entity-media-type entity) 'message)
+	       (if (and entity
+			(eq (mime-entity-media-type entity) 'message)
 			(eq (mime-entity-media-subtype entity) 'rfc822))
 		   (car (mime-entity-children entity))
 		 entity)))
@@ -100,8 +147,7 @@ It calls following-method selected from variable
 			(mime-insert-header current-entity fields)
 			t))
 	    (setq fields (std11-collect-field-names)
-		  current-entity (mime-entity-parent current-entity))
-	    ))
+		  current-entity (mime-entity-parent current-entity))))
 	(let ((rest mime-view-following-required-fields-list)
 	      field-name ret)
 	  (while rest
@@ -119,26 +165,153 @@ It calls following-method selected from variable
 						   entity field-name))))
 			(setq entity (mime-entity-parent entity)))))
 		  (if ret
-		      (insert (concat field-name ": " ret "\n"))
-		    )))
-	    (setq rest (cdr rest))
-	    ))
-	)
+		      (insert (concat field-name ": " ret "\n")))))
+	    (setq rest (cdr rest)))))
       (let ((f (cdr (assq mode mime-preview-following-method-alist))))
 	(if (functionp f)
 	    (funcall f new-buf)
 	  (message
 	   "Sorry, following method for %s is not implemented yet."
-	   mode)
-	  ))
-      )))
+	   mode))))))
 
 (defalias 'wl-draft-enclose-digest-region 'mime-edit-enclose-digest-region)
 
+(defun wl-draft-preview-attributes-list ()
+  (if (listp (car wl-draft-preview-attributes-list))
+      (elmo-uniq-list
+       (append (and (wl-message-mail-p)
+		    (cdr (assq 'mail wl-draft-preview-attributes-list)))
+	       (and (wl-message-news-p)
+		    (cdr (assq 'news wl-draft-preview-attributes-list)))))
+    wl-draft-preview-attributes-list))
+
+(defun wl-draft-show-attributes-buffer (attribute-values)
+  (let* ((cur-win (selected-window))
+	 (size (min
+		(- (window-height cur-win)
+		   window-min-height 1)
+		(- (window-height cur-win)
+		   (max
+		    window-min-height
+		    (1+ wl-draft-preview-attributes-buffer-lines))))))
+    (split-window cur-win (if (> size 0) size window-min-height))
+    (select-window (next-window))
+    (let ((pop-up-windows nil))
+      (switch-to-buffer (get-buffer-create
+			 wl-draft-preview-attributes-buffer-name)))
+    (with-current-buffer
+	(get-buffer wl-draft-preview-attributes-buffer-name)
+      (setq buffer-read-only t)
+      (let (buffer-read-only)
+	(erase-buffer)
+	(dolist (pair attribute-values)
+	  (insert (capitalize (symbol-name (car pair))) ": "
+		  (format "%s" (or (cdr pair) ""))
+		  "\n"))
+	(goto-char (point-min))
+	(wl-highlight-headers)))
+    (select-window cur-win)))
+
+(defun wl-draft-hide-attributes-buffer ()
+  (let (window buffer)
+    (when (setq window (get-buffer-window
+			wl-draft-preview-attributes-buffer-name))
+      (select-window window)
+      (delete-window))
+    (when (setq buffer (get-buffer wl-draft-preview-attributes-buffer-name))
+      (kill-buffer buffer))))
+
+(defun wl-draft-attribute-recipients ()
+  (concat (mapconcat
+	   'identity
+	   (wl-draft-deduce-address-list
+	    (current-buffer)
+	    (point-min)
+	    (save-excursion
+	      (goto-char (point-min))
+	      (re-search-forward
+	       (concat
+		"^"
+		(regexp-quote mail-header-separator)
+		"$")
+	       nil t)
+	      (point)))
+	   ", ")))
+
+(defun wl-draft-attribute-envelope-from ()
+  (or wl-envelope-from
+      (wl-address-header-extract-address wl-from)))
+
+(defun wl-draft-attribute-smtp-posting-server ()
+  (or wl-smtp-posting-server
+      (progn (require 'smtp) smtp-server)
+      "localhost"))
+
+(defun wl-draft-attribute-smtp-posting-port ()
+  (or wl-smtp-posting-port
+      (progn (require 'smtp) smtp-service)))
+
+(defun wl-draft-attribute-newsgroups ()
+  (std11-field-body "Newsgroups"))
+
+(defun wl-draft-nntp-attribute (attribute &optional alternatives)
+  (let ((config (cdr (elmo-string-matched-assoc
+		      (std11-field-body "newsgroups")
+		      wl-nntp-posting-config-alist)))
+	entry)
+    (when (stringp config)
+      (setq config (list (cons 'server config))))
+    (if (setq entry (assq attribute config))
+	;; maybe nil
+	(cdr entry)
+      (let (value)
+	(while alternatives
+	  (if (setq value (symbol-value (car alternatives)))
+	      (setq alternatives nil)
+	    (setq alternatives (cdr alternatives))))
+	value))))
+
+(defun wl-draft-attribute-nntp-posting-server ()
+  (wl-draft-nntp-attribute
+   'server
+   '(wl-nntp-posting-server elmo-nntp-default-server)))
+
+(defun wl-draft-attribute-nntp-posting-port ()
+  (wl-draft-nntp-attribute
+   'point
+   '(wl-nntp-posting-port elmo-nntp-default-port)))
+
+(defun wl-draft-attribute-value (attr)
+  (let ((name (symbol-name attr))
+	fsymbol symbol)
+    (cond ((and (setq fsymbol (intern-soft
+			       (format "wl-draft-attribute-%s" name)))
+		(fboundp fsymbol))
+	   (funcall fsymbol))
+	  ((and (setq symbol (intern-soft (format "wl-%s" name)))
+		(boundp symbol))
+	   (symbol-value symbol))
+	  ((boundp attr)
+	   (symbol-value attr)))))
+
+(defun wl-mime-quit-preview ()
+  "Quitting method for mime-view."
+  (let* ((temp (and (boundp 'mime-edit-temp-message-buffer) ;; for SEMI <= 1.14.6
+		    mime-edit-temp-message-buffer))
+	 (window (selected-window))
+	 buf)
+    (mime-preview-kill-buffer)
+    (set-buffer temp)
+    (setq buf mime-edit-buffer)
+    (kill-buffer temp)
+    (select-window window)
+    (switch-to-buffer buf)))
+
 (defun wl-draft-preview-message ()
   "Preview editing message."
   (interactive)
-  (let* (recipients-message
+  (let* (attribute-values
+	 (orig-buffer (current-buffer))
 	 (current-point (point))
 	 (config-exec-flag wl-draft-config-exec-flag)
 	 (parent-folder wl-draft-parent-folder)
@@ -151,42 +324,46 @@ It calls following-method selected from variable
 	       (symbol-value 'mime-header-encode-method-alist))))
 	 mime-view-ignored-field-list	; all header.
 	 (mime-edit-translate-buffer-hook
-	  (append
-	   (list
-	    (function
-	     (lambda ()
-	       (let ((wl-draft-config-exec-flag config-exec-flag)
-		     (wl-draft-parent-folder parent-folder))
-		 (goto-char current-point)
-		 (run-hooks 'wl-draft-send-hook)
-		 (setq recipients-message
-		       (condition-case err
-			   (concat "Recipients: "
-				   (mapconcat
-				    'identity
-				    (wl-draft-deduce-address-list
-				     (current-buffer)
-				     (point-min)
-				     (save-excursion
-				       (goto-char (point-min))
-				       (re-search-forward
-					(concat
-					 "^"
-					 (regexp-quote mail-header-separator)
-					 "$")
-					nil t)
-				       (point)))
-				    ", "))
-			 (error
-			  (kill-buffer (current-buffer))
-			  (signal (car err) (cdr err)))))))))
+	  (cons
+	   (lambda ()
+	     (let ((wl-draft-config-exec-flag config-exec-flag)
+		   (wl-draft-parent-folder parent-folder)
+		   (copy-buffer (current-buffer)))
+	       (wl-copy-local-variables
+		orig-buffer
+		copy-buffer
+		(with-current-buffer orig-buffer
+		  (append wl-draft-config-variables
+			  (wl-draft-clone-local-variables))))
+	       (goto-char current-point)
+	       (run-hooks 'wl-draft-send-hook)
+	       (condition-case err
+		   (setq attribute-values
+			 (mapcar
+			  (lambda (attr)
+			    (cons attr (wl-draft-attribute-value attr)))
+			  (if wl-draft-preview-attributes
+			      (wl-draft-preview-attributes-list)
+			    '(recipients))))
+		 (error
+		  (kill-buffer (current-buffer))
+		  (signal (car err) (cdr err))))))
 	   mime-edit-translate-buffer-hook)))
     (mime-edit-preview-message)
+    (make-local-variable 'mime-preview-quitting-method-alist)
+    (setq mime-preview-quitting-method-alist
+	  '((mime-temp-message-mode . wl-mime-quit-preview)))
     (let ((buffer-read-only nil))
       (when wl-highlight-body-too
 	(wl-highlight-body))
       (run-hooks 'wl-draft-preview-message-hook))
-    (message "%s" recipients-message)))
+    (make-local-hook 'kill-buffer-hook)
+    (add-hook 'kill-buffer-hook #'wl-draft-hide-attributes-buffer nil t)
+    (if wl-draft-preview-attributes
+	(ignore-errors ; in case when the window is too small
+	 (wl-draft-show-attributes-buffer attribute-values))
+      (message (concat "Recipients: "
+		       (cdr (assq 'recipients attribute-values)))))))
 
 (defalias 'wl-draft-caesar-region  'mule-caesar-region)
 
@@ -219,8 +396,6 @@ It calls following-method selected from variable
    (elmo-message-fetch (wl-folder-get-elmo-folder folder)
 		       number
 		       (elmo-make-fetch-strategy 'entire)
-		       nil
-		       (current-buffer)
 		       'unread)
    (mime-parse-buffer nil)))
 
@@ -247,6 +422,7 @@ It calls following-method selected from variable
     (widen)
     (let* ((entity (get-text-property (point) 'mime-view-entity))
 	   (node-id (mime-entity-node-id entity))
+	   (filename (mime-entity-safe-filename entity))
 	   (header-start (mime-buffer-entity-header-start-internal entity))
 	   (body-end (mime-buffer-entity-body-end-internal entity))
 	   (folder (wl-folder-get-elmo-folder wl-message-buffer-cur-folder))
@@ -258,8 +434,9 @@ It calls following-method selected from variable
       (with-current-buffer orig-buf
 	(unless (string-equal
 		 (buffer-string)
-		 (elmo-message-fetch folder number
-				     (elmo-make-fetch-strategy 'entire)))
+		 (elmo-message-fetch-string
+		  folder number
+		  (elmo-make-fetch-strategy 'entire)))
 	  (error "Buffer content differs from actual message")))
       (when (and (elmo-folder-writable-p folder)
 		 (buffer-live-p orig-buf)
@@ -267,25 +444,139 @@ It calls following-method selected from variable
 		 (yes-or-no-p
 		  (format "Do you really want to delete part %s? "
 			  (wl-mime-node-id-to-string node-id))))
-	(with-temp-buffer
-	  (insert-buffer orig-buf)
-	  (delete-region header-start body-end)
-	  (goto-char header-start)
-	  (insert "Content-Type: text/plain; charset=US-ASCII\n\n")
-	  (insert "** This part has been removed by Wanderlust **\n\n")
-	  (elmo-folder-append-buffer folder t))
+	(when (with-temp-buffer
+		(insert-buffer orig-buf)
+		(delete-region header-start body-end)
+		(goto-char header-start)
+		(insert "Content-Type: text/plain; charset=US-ASCII\n")
+		(when filename
+		  (insert
+		   "Content-Disposition:"
+		   (mime-encode-field-body
+			 (concat ""
+				 (and filename
+				      (concat " inline; filename=" (std11-wrap-as-quoted-string filename))))
+			 "Content-Disposition")
+		   "\n"))
+		(insert "\n")
+		(insert "** This part has been removed by Wanderlust **\n\n")
+		(elmo-folder-append-buffer folder))
 
-	(elmo-folder-append-messages
-	 (wl-folder-get-elmo-folder wl-trash-folder)
-	 folder (list number) nil)
-	(elmo-folder-delete-messages folder (list number))
+	  (elmo-folder-move-messages
+	   folder (list number)
+	   (wl-folder-get-elmo-folder wl-trash-folder))
+	  (when (and (elmo-cache-get-path msgid)
+		     (file-exists-p (elmo-cache-get-path msgid)))
+	    (delete-file (elmo-cache-get-path msgid)))
 
-	(when (file-exists-p (elmo-cache-get-path msgid))
-	  (delete-file (elmo-cache-get-path msgid)))
+	  (mime-preview-quit)
+	  (wl-summary-delete-messages-on-buffer (list number))
+	  (wl-summary-toggle-disp-msg 'off)
+	  (setq wl-message-buffer nil)
+	  (wl-summary-sync nil "update"))))))
 
-	(mime-preview-quit)
-	(wl-summary-toggle-disp-msg 'off)
-	(wl-summary-sync nil "update")))))
+(defun wl-message-decrypt-pgp-nonmime ()
+  "Decrypt PGP encrypted region"
+  (interactive)
+  (require 'pgg)
+  (save-excursion
+    (beginning-of-line)
+    (if (or (re-search-forward "^-+END PGP MESSAGE-+$" nil t)
+	    (re-search-backward "^-+END PGP MESSAGE-+$" nil t))
+	(let (beg end status)
+	  (setq end (match-end 0))
+	  (if (setq beg (re-search-backward "^-+BEGIN PGP MESSAGE-+$" nil t))
+	      (let ((inhibit-read-only t)
+		    (buffer-file-coding-system wl-cs-autoconv))
+		(setq status (pgg-decrypt-region beg end))
+		(pgg-display-output-buffer beg end status))
+	    (message "Cannot find pgp encrypted region")))
+      (message "Cannot find pgp encrypted region"))))
+
+(defun wl-message-verify-pgp-nonmime ()
+  "Verify PGP signed region"
+  (interactive)
+  (require 'pgg)
+  (save-excursion
+    (beginning-of-line)
+    (if (and (or (re-search-forward "^-+END PGP SIGNATURE-+$" nil t)
+		 (re-search-backward "^-+END PGP SIGNATURE-+$" nil t))
+	     (re-search-backward "^-+BEGIN PGP SIGNED MESSAGE-+$" nil t))
+	(let (status)
+	  (let* ((beg (point))
+		 (situation (mime-preview-find-boundary-info))
+		 (p-end (aref situation 1))
+		 (entity (aref situation 2))
+		 (count 0))
+	    (goto-char p-end)
+	    (while (< beg (point))
+	      (if (re-search-backward "^-+BEGIN PGP SIGNED MESSAGE-+$" nil t)
+		  (setq count (+ count 1))
+		(debug)))
+	    (with-temp-buffer
+	      (set-buffer-multibyte nil)
+	      (insert (mime-entity-body entity))
+	      (goto-char (point-max))
+	      (while (> count 0)
+		(if (re-search-backward "^-+BEGIN PGP SIGNED MESSAGE-+$" nil t)
+		    (setq count (- count 1))
+		  (debug)))
+	      (let ((r-beg (point))
+		    (r-end (re-search-forward "^-+END PGP SIGNATURE-+$" nil t)))
+		(if r-end
+		    (setq status (pgg-verify-region r-beg r-end nil 'fetch))
+		  (debug)))))
+	  (mime-show-echo-buffer)
+	  (set-buffer mime-echo-buffer-name)
+	  (set-window-start
+	   (get-buffer-window mime-echo-buffer-name)
+	   (point-max))
+	  (insert-buffer-substring
+	   (if status pgg-output-buffer pgg-errors-buffer)))
+      (message "Cannot find pgp signed region"))))
+
+;; XXX: encrypted multipart isn't represented as multipart
+(defun wl-mime-preview-application/pgp (parent-entity entity situation)
+  (require 'pgg)
+  (goto-char (point-max))
+  (let ((p (point))
+	raw-buf to-buf representation-type child-entity)
+    (goto-char p)
+    (save-restriction
+      (narrow-to-region p p)
+      (setq to-buf (current-buffer))
+      (with-temp-buffer
+	(setq raw-buf (current-buffer))
+	(mime-insert-entity entity)
+	(when (progn
+		(goto-char (point-min))
+		(re-search-forward "^-+BEGIN PGP MESSAGE-+$" nil t))
+	  (pgg-decrypt-region (point-min)(point-max))
+	  (delete-region (point-min) (point-max))
+	  (insert-buffer pgg-output-buffer)
+	  (setq representation-type 'elmo-buffer))
+	(setq child-entity (mime-parse-message
+			    (mm-expand-class-name representation-type)
+			    nil
+			    parent-entity
+			    (mime-entity-node-id-internal parent-entity)))
+	(mime-display-entity
+	 child-entity
+	 nil
+	 `((header . visible)
+	   (body . visible)
+	   (entity-button . invisible))
+	 to-buf)))))
+
+(defun wl-mime-preview-application/pgp-encrypted (entity situation)
+  (let* ((entity-node-id (mime-entity-node-id entity))
+	 (mother (mime-entity-parent entity))
+	 (knum (car entity-node-id))
+	 (onum (if (> knum 0)
+		   (1- knum)
+		 (1+ knum)))
+	 (orig-entity (nth onum (mime-entity-children mother))))
+    (wl-mime-preview-application/pgp entity orig-entity situation)))
 
 ;;; Summary
 (defun wl-summary-burst-subr (message-entity target number)
@@ -305,9 +596,7 @@ It calls following-method selected from variable
 		 (car (mime-entity-children message-entity)))
 	   (with-temp-buffer
 	     (insert (mime-entity-body message-entity))
-	     (elmo-folder-append-buffer
-	      target
-	      (mime-entity-fetch-field entity "Message-ID")))))
+	     (elmo-folder-append-buffer target))))
     number))
 
 (defun wl-summary-burst (&optional arg)
@@ -316,15 +605,23 @@ With ARG, ask destination folder."
   (interactive "P")
   (let ((raw-buf (wl-summary-get-original-buffer))
 	(view-buf wl-message-buffer)
-	children message-entity content-type target)
+	message-entity target)
     (save-excursion
-      (setq target wl-summary-buffer-elmo-folder)
-      (when (or arg (not (elmo-folder-writable-p target)))
-	(let ((fld (wl-summary-read-folder wl-default-folder "to extract to")))
-	  (setq target (wl-folder-get-elmo-folder fld))))
+      (when (and (null arg)
+		 (elmo-folder-writable-p wl-summary-buffer-elmo-folder))
+	(setq target wl-summary-buffer-elmo-folder))
+      (while (null target)
+	(let ((name (wl-summary-read-folder wl-default-folder
+					    "to extract to")))
+	  (setq target (wl-folder-get-elmo-folder name))
+	  (unless (elmo-folder-writable-p target)
+	    (message "%s is not writable" name)
+	    (setq target nil)
+	    (sit-for 1))))
       (wl-summary-set-message-buffer-or-redisplay)
       (with-current-buffer view-buf
-	(setq message-entity (get-text-property (point-min) 'mime-view-entity)))
+	(setq message-entity
+	      (get-text-property (point-min) 'mime-view-entity)))
       (when message-entity
 	(message "Bursting...")
 	(with-current-buffer raw-buf
@@ -332,18 +629,22 @@ With ARG, ask destination folder."
 	(message "Bursting...done"))
       (if (elmo-folder-plugged-p target)
 	  (elmo-folder-check target)))
-    (wl-summary-sync-update)))
+    (when (and target
+	       (string= wl-summary-buffer-folder-name
+			(elmo-folder-name-internal target)))
+      (save-excursion (wl-summary-sync-update)))))
 
 ;; internal variable.
 (defvar wl-mime-save-directory nil "Last saved directory.")
 ;;; Yet another save method.
 (defun wl-mime-save-content (entity situation)
-  (let ((filename (read-file-name "Save to file: "
-				  (expand-file-name
-				   (or (mime-entity-safe-filename entity)
-				       ".")
-				   (or wl-mime-save-directory
-				       wl-temporary-file-directory)))))
+  (let ((filename (expand-file-name
+		   (read-file-name "Save to file: "
+				   (expand-file-name
+				    (or (mime-entity-safe-filename entity)
+					".")
+				    (or wl-mime-save-directory
+					wl-temporary-file-directory))))))
     (while (file-directory-p filename)
       (setq filename (read-file-name "Please set filename (not directory): "
 				     filename)))
@@ -358,11 +659,9 @@ With ARG, ask destination folder."
 (defun wl-mime-combine-message/partial-pieces (entity situation)
   "Internal method for wl to combine message/partial messages automatically."
   (interactive)
-  (let* ((msgdb (save-excursion
-		  (set-buffer wl-message-buffer-cur-summary-buffer)
-		  (wl-summary-buffer-msgdb)))
+  (let* ((folder (with-current-buffer wl-message-buffer-cur-summary-buffer
+		   wl-summary-buffer-elmo-folder))
 	 (mime-display-header-hook 'wl-highlight-headers)
-	 (folder wl-message-buffer-cur-folder)
 	 (id (or (cdr (assoc "id" situation)) ""))
 	 (mother (current-buffer))
 	 (summary-buf wl-message-buffer-cur-summary-buffer)
@@ -393,18 +692,16 @@ With ARG, ask destination folder."
 	      wl-summary-buffer-mime-charset)))
       (if (string-match "[0-9\n]+" subject-id)
 	  (setq subject-id (substring subject-id 0 (match-beginning 0))))
-      (setq overviews (elmo-msgdb-get-overview msgdb))
       (catch 'tag
-	(while overviews
+	(elmo-folder-do-each-message-entity (entity folder)
 	  (when (string-match
 		 (regexp-quote subject-id)
-		 (elmo-msgdb-overview-entity-get-subject (car overviews)))
+		 (elmo-message-entity-field entity 'subject 'decode))
 	    (let* ((message
 		    ;; request message at the cursor in Subject buffer.
 		    (wl-message-request-partial
-		     folder
-		     (elmo-msgdb-overview-entity-get-number
-		      (car overviews))))
+		     (elmo-folder-name-internal folder)
+		     (elmo-message-entity-number entity)))
 		   (situation (mime-entity-situation message))
 		   (the-id (or (cdr (assoc "id" situation)) "")))
 	      (when (string= (downcase the-id)
@@ -412,8 +709,7 @@ With ARG, ask destination folder."
 		(with-current-buffer mother
 		  (mime-store-message/partial-piece message situation))
 		(if (file-exists-p full-file)
-		    (throw 'tag nil)))))
-	  (setq overviews (cdr overviews)))
+		    (throw 'tag nil))))))
 	(message "Not all partials found.")))))
 
 (defun wl-mime-display-text/plain (entity situation)
@@ -423,7 +719,7 @@ With ARG, ask destination folder."
 
 (defun wl-mime-display-header (entity situation)
   (let ((elmo-message-ignored-field-list
-	 (if wl-message-buffer-all-header-flag
+	 (if wl-message-buffer-require-all-header
 	     nil
 	   wl-message-ignored-field-list))
 	(elmo-message-visible-field-list wl-message-visible-field-list)
@@ -437,7 +733,7 @@ With ARG, ask destination folder."
     (mime-decrypt-application/pgp-encrypted entity situation)
     (setq wl-message-buffer-cur-summary-buffer summary-buffer)
     (setq wl-message-buffer-original-buffer original-buffer)))
-   
+
 
 ;;; Setup methods.
 (defun wl-mime-setup ()
@@ -461,6 +757,14 @@ With ARG, ask destination folder."
      (body-presentation-method . wl-mime-display-text/plain)
      (major-mode . wl-original-message-mode)))
 
+  (ctree-set-calist-strictly
+   'mime-preview-condition
+   '((type . application)(subtype . pgp-encrypted)
+     (encoding . t)
+     (body . invisible)
+     (body-presentation-method . wl-mime-preview-application/pgp-encrypted)
+     (major-mode . wl-original-message-mode)))
+
   (ctree-set-calist-strictly
    'mime-acting-condition
    '((type . message) (subtype . partial)
